"use strict";

const PORT_NUM = 2903;

const express = require("express");
const app = express();

const sqlite3 = require("sqlite3");
const sqlite = require("sqlite");

const multer = require("multer");

app.use(express.urlencoded({extended: true}));
app.use(express.json());
app.use(multer().none());

const validRating = /[12345]{1}/;
const validIsAnon = /[01]{1}/;

// create reviews database
let reviewsDB = new sqlite3.Database('./reviews.db', sqlite3.OPEN_READWRITE, (e) => {
    // check if database exists yet
    if (e && e.code == "SQLITE_CANTOPEN") {
        // database does not exist, so we create one
        var newDB = new sqlite3.Database('./reviews.db', (e) => {
            if (e) {
                console.log("Can't create a new database " + e);
                exit(1);
            }
            setupReviewsDB(newDB);
        })
        return;
    } else if (e) {
        console.log("Database exists but there is another error: " + e);
        exit(1);
    }
});

function setupReviewsDB(newDB) {
    newDB.exec(`
    CREATE TABLE "classes" (
        "id"	INTEGER NOT NULL,
        "name"	TEXT NOT NULL,
        "description"	TEXT,
        PRIMARY KEY("id" AUTOINCREMENT)
    );

    CREATE TABLE "reviews" (
        "review_id"	INTEGER NOT NULL,
        "difficulty_rating"	INTEGER NOT NULL,
        "time_rating"	INTEGER NOT NULL,
        "fun_rating"	INTEGER NOT NULL,
        "overall_rating"	INTEGER NOT NULL,
        "review_text"	TEXT NOT NULL,
        "user"	TEXT,
        "class"	TEXT,
        "is_anon"	INTEGER NOT NULL,
        PRIMARY KEY("review_id")
    );

    CREATE TABLE "users" (
        "username"	TEXT NOT NULL,
        "displayname"	TEXT NOT NULL,
        PRIMARY KEY("username")
    );

    `, () => {
        console.log("Setup reviews database!");
    });
}

// test endpoint
app.get("/test", async function(req, res) {
    res.type("text").send("Hello world.");
});

// endpoint for getting all reviews from a class with the class id (E.g. /reviews/CSE 403)
app.get("/reviews/:classid", async function(req, res) {
    try {
        let className = req.params.classid;
        let db = await getDBConnection();
        let qry = "SELECT * FROM reviews INNER JOIN users ON reviews.user = users.username WHERE class = ?;";
        let reviews = await db.all(qry, className);

        reviews.forEach(function (review, index) {
            // if review is stored as anonymous, anonymize real name
            // before sending GET result to user's frontend for security
            if (review.is_anon === 1) {
                review.displayname = 'Anonymous User';
            }
        })

        res.json(reviews);

        db.close();
    } catch (err) {
        res.type("text");
        res.status(500).send("An error occured on the server.");
    }
});

// endpoint for getting average ratings from all reviews of a class with a class id
app.get("/reviews/:classid/ratings", async function (req, res) {
    try {
        let className = req.params.classid;
        let db = await getDBConnection();

        let qry = "SELECT AVG(difficulty_rating) AS avg_difficulty," +
                  " AVG(time_rating) AS avg_time, AVG(fun_rating) AS avg_fun," +
                  " AVG(overall_rating) AS avg_overall FROM reviews WHERE class = ?;";
        let ratings = await db.all(qry, className);

        res.json(ratings);

        db.close();
    } catch (err) {
        res.type("text");
        res.status(500).send("An error occured on the server.");
    }
});

// endpoint for getting all reviews from a given user (E.g. /reviews/user/klee33)
app.get("/reviews/user/:username", async function (req, res) {
    try {
        let username = req.params.username;
        let db = await getDBConnection();
        let qry = "SELECT * FROM reviews INNER JOIN users ON reviews.user = users.username WHERE user = ?;";
        let reviews = await db.all(qry, username);

        res.json(reviews);

        db.close();
    } catch (err) {
        res.type("text");
        res.status(500).send("An error occured on the server.");
    }
});

// I don't believe this endpoint is necessary, but its here just in case

/*app.get("/reviews", async function(req, res) {
    try {
        let className = req.params.classid;
        let db = await getDBConnection();

        let classQry = "SELECT * FROM classes WHERE "

        let qry = "SELECT * FROM reviews;";
        let reviews = await db.all(qry, className);

        res.json(reviews);

        db.close();
    } catch (err) {
        res.type("text");
        res.status(500).send("An error occured on the server.");
    }
});*/

// endpoint for posting a review to the database for a given class and user
app.post("/reviews/post", async function(req, res) {
    res.type("text");
    try {
        let difficultyRating = req.body["difficulty-rating"];
        let timeRating = req.body["time-rating"];
        let funRating = req.body["fun-rating"];
        let overallRating = req.body["overall-rating"];
        let reviewText = req.body["review-text"];
        let className = req.body["class-name"];
        let username = req.body["username"];
        let isAnon = req.body["is-anon"];
        if (isAnon === undefined) {
            isAnon = '0';
        } else {
            isAnon = '1';
        }
        // should this line above be:
        // let isAnon = req.body["is-anon"].toString();
        let db = await getDBConnection();
        let isDuplicate = await isDuplicateForUser(className, username, db);

        if (difficultyRating === undefined || timeRating === undefined
            || funRating === undefined || overallRating === undefined
            || reviewText === undefined || className === undefined
            || username === undefined || isAnon === undefined) {
            res.status(400).send("One or more parameters missing.");
        } else if (!ratingIsValid(difficultyRating) || !ratingIsValid(timeRating)
                    || !ratingIsValid(funRating) || !ratingIsValid(overallRating)) {
            res.status(400).send("One or more ratings are invalid. (Must be integers between 1-5)");
        } else if (!isAnonIsValid(isAnon)) {
            res.status(400).send("is-anon POST parameter must be 0 or 1.")
        } else if (!(await userExists(username, db))) {
            res.status(400).send("Invalid username given; user does not exist");
        } else if (isDuplicate !== null) {
            res.status(400).send("User has already left a review for this class: " + isDuplicate);
        } else {
            let classExistsQry = "SELECT * FROM classes WHERE name = ?;";
            let classExists = await db.all(classExistsQry, className);

            // If class doesn't exist in database add it to database
            if (classExists.length === 0) {
                let addClassQry = "INSERT INTO classes (name, description) VALUES (?, null);";
                await db.run(addClassQry, className);
            }

            let addReviewQry = "INSERT INTO reviews (difficulty_rating, time_rating, fun_rating, overall_rating," +
                                " review_text, user, class, is_anon) VALUES (?, ?, ?, ?, ?, ?, ?, ?);";
            await db.run(addReviewQry, difficultyRating, timeRating, funRating, overallRating, reviewText,
                          username, className, isAnon);
            res.send("Review successfully added.");
        }
        db.close();
    } catch (err) {
        console.log(err);
        res.status(500).send("An error occured on the server.");
    }
});

// endpoint for updating an existing review to change description, ratings, and anonymity
app.post("/reviews/update", async function(req, res) {
    res.type("text");
    try {
        let difficultyRating = req.body["difficulty-rating"];
        let timeRating = req.body["time-rating"];
        let funRating = req.body["fun-rating"];
        let overallRating = req.body["overall-rating"];
        let reviewText = req.body["review-text"];
        let isAnon = req.body["is-anon"];
        if (isAnon === undefined) {
            isAnon = '0';
        } else {
            isAnon = '1';
        }

        let reviewID = req.body["review-id"];

        let db = await getDBConnection();

        if (difficultyRating === undefined || timeRating === undefined
            || funRating === undefined || overallRating === undefined
            || reviewText === undefined || isAnon === undefined
            || reviewID === undefined) {
            res.status(400).send("One or more parameters missing.");
        } else if (!ratingIsValid(difficultyRating) || !ratingIsValid(timeRating)
                    || !ratingIsValid(funRating) || !ratingIsValid(overallRating)) {
            res.status(400).send("One or more ratings are invalid. (Must be integers between 1-5)");
        } else if (!isAnonIsValid(isAnon)) {
            res.status(400).send("is-anon POST parameter must be 0 or 1.");
        } else if (!(await reviewExists(reviewID, db))) {
            res.status(400).send("Invalid review to update.");
        } else {
            let updateReviewQry = "UPDATE reviews SET difficulty_rating = ?, time_rating = ?, fun_rating = ?, " +
                                  "overall_rating = ?, review_text = ?, is_anon = ? WHERE review_id = ?;";
            await db.run(updateReviewQry, difficultyRating, timeRating, funRating, overallRating, reviewText,
                         isAnon, reviewID);
            res.send("Review successfully updated.");
        }
        db.close();
    } catch (err) {
        res.status(500).send("An error occured on the server.");
    }
});

// endpoint for deleting a review
app.post("/reviews/delete", async function(req, res) {
    res.type("text");
    try {
        let reviewID = req.body["review-id"];

        let db = await getDBConnection();

        if (reviewID === undefined) {
            res.status(400).send("One or more parameters missing.");
        } else if (!(await reviewExists(reviewID, db))) {
            res.status(400).send("Invalid review to delete.");
        } else {
            let deleteReviewQry = "DELETE FROM reviews WHERE review_id = ?;";
            await db.run(deleteReviewQry, reviewID);
            res.send("Review successfully deleted.");
        }
    } catch (err) {
        res.status(500).send("An error occured on the server.");
    }
});

// endpoint for creating a new user
app.post("/users/create", async function(req, res) {
    res.type("text");
    try {
        let username = req.body["username"];
        let displayName = req.body["display-name"];

        let db = await getDBConnection();

        if (username === undefined || displayName === undefined) {
            res.status(400).send("One or more parameters missing.");
        } else if (await userExists(username, db)) {
            res.status(400).send("Username is already taken.");
        } else {
            let addUserQry = "INSERT INTO users (username, displayname) VALUES (?, ?);";
            await db.run(addUserQry, username, displayName);
            res.send("User successfully added.");
        }
        db.close();
    } catch (err) {
        res.status(500).send("An error occured on the server.");
    }
});

// endpoint for logging in as a user
app.post("/users/login", async function(req, res) {
    res.type("text");
    try {
        let username = req.body["username"];

        let db = await getDBConnection();

        if (username === undefined) {
            res.status(400).send("One or more parameters missing.");
        } else if (!(await userExists(username, db))) {
            res.status(400).send("Invalid login creditials")
        } else {
            res.send("Login success.");
        }
        db.close();
    } catch (err) {
        res.status(500).send("An error occured on the server.");
    }
});

// May be necessary later if we want to use cookies to store a login session
/*app.post("/users/logout", async function(req, res) {
    res.type("text");
    res.send("Logout success.");
});*/

/**
 * Connects to the store database and returns it
 * @returns {object} - Database object
 */
async function getDBConnection() {
    const db = await sqlite.open({
        filename: "reviews.db",
        driver: sqlite3.Database
    });
    return db;
}

/**
 * Given a rating returns whether it is valid or not
 * @param {string} rating - Rating to test
 * @returns {boolean} - Whether or not the rating is valid
 */
function ratingIsValid(rating) {
    let match = rating.match(validRating);
    return match != null && match.length == 1 && match[0].length == rating.length;
}

/**
 * Given an isAnon post parameter returns whether it is valid or not
 * @param {string} isAnon - Post parameter to test
 * @returns {boolean} - Whether or not the it is valid
 */
function isAnonIsValid(isAnon) {
    let match = isAnon.match(validIsAnon);
    return match != null && match.length == 1 && match[0].length == isAnon.length;
}

/**
 * Checks to see if a user exists in the database
 * @param {string} username - Username to check
 * @param {object} db - Database instance to use
 * @returns {boolean} - If the user exists in the database
 */
async function userExists(username, db) {
    let userExistsQry = "SELECT * FROM users WHERE username = ?;";
    let userExists = await db.all(userExistsQry, username);
    return userExists.length > 0;
}

/**
 * Checks to see if a review exists in the database
 * @param {string} review - Review ID to check
 * @param {object} db - Database instance to use
 * @returns {boolean} - If the review exists in the database
 */
async function reviewExists(review, db) {
    let reviewExistsQry = "SELECT * FROM reviews WHERE review_id = ?;";
    let reviewExists = await db.all(reviewExistsQry, review);
    return reviewExists.length > 0;
}

/**
 * Checks to see if a review exists in the database given a class and a user
 * @param {string} classname - Class ID to check
 * @param {string} username - Username to check
 * @param {object} db - Database instance to use
 * @returns {int} - ReviewID if it exists in the database, null otherwise
 */
async function isDuplicateForUser(classname, username, db)  {
    let reviewExistsQry = "SELECT review_id FROM reviews WHERE class = ? AND user = ?;";
    let reviewExists = await db.all(reviewExistsQry, classname, username);
    if (reviewExists.length > 0) {
        return reviewExists[0]["review_id"];
    }
    return null;
}

app.use(express.static("public"));
const PORT = process.env.PORT || PORT_NUM;
app.listen(PORT);

module.exports = app;