const LOCAL_DEBUG = false;
const url = (LOCAL_DEBUG) ? "http://localhost:2903" : "http://attu7.cs.washington.edu:2903";

let currUser = "";
const validClassCode = /^\d{3}$/;
let currClassCode = null;

/**
 * Creates a slider component with a label and displays its current value.
 * @param {string} id - The ID attribute for the wrapping div element.
 * @param {string} labelText - The text content of the label element.
 * @param {string} name - The name attribute for the input element.
 * @returns {string} HTML string for the slider component.
 */
function createSliderComponent(id, labelText, name) {
  // Return the HTML template string for the slider component
  return `
    <div id="${id}">
      <label for="${name}">${labelText}:</label>
      <div class="slider-container">
        <input type="range" id="${name}" name="${name}" min="1" max="5">
        <span id="${name}_value" class="slider-value">3</span>
      </div>
    </div>
  `;
}


/**
 * Sets up the sliders on the page and adds event listeners to update the displayed value when the user interacts with the sliders.
 */
function updateValue(slider) {
  // Set the text content of the element with the ID "{slider.id}_value" to the current slider value
  document.getElementById(`${slider.id}_value`).textContent = slider.value;
}
// Here is the new function for form submission
function handleSubmit(e) {
  e.preventDefault();

  let formData = new FormData(e.target);
  console.log(formData);

  formData.append("username", currUser);
  if (currClassCode === null) {
    alert("Not on a valid class page!");
    return;
  }
  formData.append("class-name", currClassCode);

  fetch(url + '/reviews/post', {method: 'POST', body: formData})
    .then(response => {
      return response.text();
    })
    .then(data => {
      if (data === 'Review successfully added.') {
        alert('Review successfully posted!');
      } else if (data.includes("User has already left a review for this class:")) {
        if (confirm("It looks like you've already left a review for this class. Would you like to overwrite your old review?")) {
          let words = data.split(" ");
          let reviewID = words[words.length - 1];
          let editFormData = formData;
          editFormData.append("review-id", reviewID);

          fetch(url + "/reviews/update", {method: "POST", body: editFormData})
            .then(resp => resp.text())
            .then(resp => {
              if (resp === "Review successfully updated.") {
                alert("Review successfully updated!");
              } else {
                alert("Error updating review: " + resp);
              }
            })
            .catch(console.error);
        }
      } else {
        alert('Error: ' + data);
      }
    })
    .catch((error) => {
      console.error('Error: ' + error);
    });
}


// Event listener for when the DOM is fully loaded
document.addEventListener('DOMContentLoaded', async function() {
  // Add the slider components for difficulty, time commitment, and fun to the element with the ID "sliders"
  document.getElementById('sliders').innerHTML = createSliderComponent('difficulty', 'Difficulty', 'difficulty-rating');
  document.getElementById('sliders').innerHTML += createSliderComponent('time', 'Time commitment', 'time-rating');
  document.getElementById('sliders').innerHTML += createSliderComponent('fun', 'Fun', 'fun-rating');
  document.getElementById('sliders').innerHTML += createSliderComponent('overall', 'Overall', 'overall-rating');
  // Event listener for the logout button
  document.getElementById('logout-button').addEventListener('click', handleLogout);

  // Get all the slider elements with input[type="range"]
  const sliders = document.querySelectorAll('input[type="range"]');
  // For each slider, add an event listener to update the displayed value when the slider is moved
  sliders.forEach(slider => {
    slider.addEventListener('input', () => updateValue(slider));
  });

  let reviewClass = document.getElementById("review-class-display");

  let tab = await getCurrentTab();
  if (isValidTab(tab.title)) {
    reviewClass.textContent = "Leaving review for: " + getClassFromTitle(tab.title);
    currClassCode = getClassFromTitle(tab.title);
  } else {
    reviewClass.textContent = "Not currently on a valid UW class page. Try reloading the page or closing and reopening this extension.";
    currClassCode = null;
  }


  document.getElementById('review-form').addEventListener('submit', handleSubmit);

  let loggedIn = document.getElementById('logged-user');
  chrome.runtime.sendMessage({type: "getCurrUser"}, (response) => {
    if (response != null) {
      currUser = response;
      loggedIn.textContent = "Logged in as: " + currUser;
    }
  });
});

// Handle logout functionality
function handleLogout() {
  currUser = ""; // Clear the current user
  document.getElementById('logged-user').textContent = "Logged in as: ";
  // Redirect to the login page
  window.location.href = "login.html";
  chrome.runtime.sendMessage({type: "logout"}, (response) => {});
}

// From https://developer.chrome.com/docs/extensions/reference/tabs/
async function getCurrentTab() {
  let queryOptions = { active: true, lastFocusedWindow: true };
  // `tab` will either be a `tabs.Tab` instance or `undefined`.
  let [tab] = await chrome.tabs.query(queryOptions);
  return tab;
}

/**
 * Given the title of the tab gets the class
 * @param {string} titleString - Tab title to check
 * @returns {string} - Valid class code for form
 */
function getClassFromTitle(titleString) {
  let words = titleString.split(" ");
  let result = "";
  for (let i = 0; i < words.length; i++) {
    result += words[i] + " ";
    if (validClassCode.test(words[i])) {
      break;
    }
  }
  return result.trim();
}

/**
 * Given the title of the tab, verifies if we are on a valid class page
 * @param {string} titleString - Tab title to check
 * @returns {boolean} - Whether or not we are on a valid class page to leave a review for
 */
function isValidTab(titleString) {
  let words = titleString.split(" ");
  // Test if the last word in the tab title is "MyPlan"
  if (words[words.length - 1] !== "MyPlan") {
    return false;
  }
  // Test if any of the words in the tab title is a valid class code number
  for (let i = 0; i < words.length; i++) {
    if (validClassCode.test(words[i])) {
      return true;
    }
  }
  return false;
}
